def reflection_permutation(n):
	string = "(1,-1)"
	string += "".join(("(%s, -%s)(%s, -%s)" % (i+2, n-i, n-i, i+2) if i+2 != n-i else "(%s,-%s)" % (i+2, i+2)) for i in range(0,int(n/2)))			
	return string
	
def rotation_permutation(n):
	positive = tuple(i for i in range(1,n+1))
	negative = tuple(-i for i in positive)
	return str(positive) + str(negative)
	
def signed_inversion(of_positions = (1,2)):
	return "not done yet, will return a signed inversion in cycles form"


from sage.all_cmdline import *

def draw_genome(n=None, permutation=None, orientation_list=None, save_as_pgf=False, show=True, name="genome", oriented=True):
	"""Produce a drawing of a circular genome with n regions"""
	# Imports
	import matplotlib
	import matplotlib.pyplot as plt
	if save_as_pgf:
		matplotlib.use("pgf")
		matplotlib.rcParams.update({
			"pgf.texsystem": "pdflatex",
			'font.family': 'serif',
			'text.usetex': True,
			'pgf.rcfonts': False
		})
	import numpy as np
	###
	
	region_labels = None
	orientation_list = None
	if permutation:
		n = len(permutation)
		region_labels = [None for _ in range(1, n+1)]
		orientation_list = [False for _ in range(1,n+1)]
		for i, r in enumerate(permutation):
			region_labels[abs(r)-1] = i+1
			orientation_list[abs(r)-1] = r < 0
		region_labels = list(reversed(region_labels))
		orientation_list = list(reversed(orientation_list))
	else:
		pass #region_labels = [k for k in range(1,n+1)]
		
	# Stuff for the pie chart
	segment_sizes = [100/n for _ in range(n)]
	segment_colors = ['white' for _ in range(n)]
	 
	fig1, ax1 = plt.subplots(figsize=(2,2))
	wedges, texts = ax1.pie(segment_sizes, labeldistance=1.2, colors=segment_colors, labels=region_labels, radius=1, startangle=90, wedgeprops={"edgecolor":"0", 'linewidth': 1, 'linestyle': 'solid', 'antialiased': True})
				
	if oriented and orientation_list:
		for i, p in enumerate(wedges):
			ang = (p.theta2 - p.theta1)/2. + p.theta1
			y = np.sin(np.deg2rad(ang))
			x = np.cos(np.deg2rad(ang))
			arrow_length = 0.00001 # We only want the arrow head
			angle = np.deg2rad(ang) + (np.pi/2 if orientation_list[i] else 3*np.pi/2)
			if orientation_list:
				ax1.arrow(x, y, arrow_length*np.cos(angle), arrow_length*np.sin(angle),overhang = 0.3, head_width=0.1, head_length=0.1, head_starts_at_zero=True, length_includes_head=True, linewidth=1, facecolor='black', fill=True)

	# Draw inner circle to hide most of the segment edges
	centre_circle = plt.Circle((0,0),0.85,fc='white')
	fig = plt.gcf()
	fig.gca().add_artist(centre_circle)
	ax1.axis('equal') # makes sure it's a circle
	plt.tight_layout()


	import os
	if not os.path.exists('_output'):
		os.makedirs('_output')
	plt.savefig(f'_output/{name}.png', facecolor='white', transparent=True)
	plt.savefig(f'_output/{name}' + ('.pgf' if save_as_pgf else '.pdf'), facecolor='white', transparent=True)
	if show and (not save_as_pgf):
		plt.show()


from enum import Enum, auto
from sage.all_cmdline import *
from cgt import conversions, helper_functions

class SET(Enum):
	SIGNED_CYCLES 	= auto()
	UNSIGNED_CYCLES = auto()
	ONE_ROW			= auto()
	WREATH			= auto()
	WREATH_S2		= auto()
	
def HyperoctahedralGroup(n, as_set_of=SET.SIGNED_CYCLES):
	### Some constants
	N 		= list(range(1, n+1))		# The set {1, ..., n}
	NN		= list(range(1, 2*n+1))		# The set {1, ..., 2n}
	pmN 	= list(range(-n, 0)) + N 	# The set {+-1, ..., +-n}
	### Symmetric groups
	S_n = SymmetricGroup(N)
	S_2n = SymmetricGroup(NN)
	S_pmn = SymmetricGroup(pmN)

	if as_set_of in [SET.SIGNED_CYCLES, SET.ONE_ROW]:
		HO = SignedPermutations(n)
		if as_set_of == SET.ONE_ROW:
			return HO
		else:
			HO_cycles = S_pmn.subgroup([
				conversions.signed_permutation_to_cycles(n, sigma) for sigma in HO.gens()
			])
			HO_cycles._as_set_of = SET.SIGNED_CYCLES
			HO_cycles._n = n
			return HO_cycles
	else:
		return "Not yet implemented"

def DihedralSubgroup(G, n=None):
	try:
		if G._as_set_of == SET.SIGNED_CYCLES:
			n = G._n
			r, s = helper_functions.rotation_permutation(n), helper_functions.reflection_permutation(n) # Generators for D_n
			D_n = G.subgroup([s,r])
			D_n._as_set_of = SET.SIGNED_CYCLES
			D_n._n = G._n
			return D_n
	except:
		print('Assuming group is S_n')
		return DihedralGroup(n)
		

def EquivalenceClasses(G, n=None, write_to_file_named=None):
	'''Example use: EquivalenceClasses(HyperoctahedralGroup(5, as_set_of=SET.SIGNED_CYCLES))'''
	H_n = G
	try:
		n = H_n._n
	except:
		print("Assuming group is S_n")
	D_n = DihedralSubgroup(H_n, n)
	cards=[]
	if write_to_file_named != None:
		sdreps = open('_output/genomes_H'+str(n)+'.txt','w')
	H_n_elements = Set(H_n)
	while H_n_elements.cardinality()>0:
		T = Set([d_1*H_n_elements[0]*d_2 for d_1 in D_n for d_2 in D_n])
		T = T.union(Set([g.inverse() for g in T]))
		H_n_elements = H_n_elements.difference(T)
		perms=[conversions.cycles_to_signed_permutation(n, str(g)) for g in T]
		#perms=sorted(perms)
		cards.append((perms[0],len(perms)))
		if write_to_file_named != None:
			sdreps.write(str(perms[0]))
			sdreps.write(str('\n'))
	if write_to_file_named != None:
		sdreps.close()
	else:
		return cards


from sage.all_cmdline import *
from cgt import hyperoctahedral_groups
import numpy as np

def regular_representation(G, g): 
	"""regular representation """
	C = QQbar # Complex field
	CG =  GroupAlgebra(G,C)
	return matrix(QQbar, [(CG(g)*G(h)).to_vector(QQbar) for h in G]).transpose()
	
def irreducible_representations(n, signed=False):
	"""Return a complete system of pairwise irreducible representations of either S_n or H_n"""
	if not signed:
		representatives = SymmetricGroupRepresentations(n).list()
		return representatives
	elif signed:
		G = hyperoctahedral_groups.HyperoctahedralGroup(n)
		characters = gap.Irr(G)
		representations = []
		for character in characters:
			irrep = gap.IrreducibleAffordingRepresentation(character)
			def representation(sigma, as_gap_matrix=False, _irrep=irrep):	
				image = gap.Image(_irrep, sigma)
				return image if as_gap_matrix else matrix(CDF, image)
			representations.append(representation)
		return representations


from sage.all_cmdline import *  # import sage library

def cycles_to_signed_permutation(n, disjoint_cycle_string):
	"""Converts a cycle notation string into two-row notation"""
	g = disjoint_cycle_string
	S = list(range(-n, 0)) + list(range(1, n+1))
	S_N = SymmetricGroup(S)
	g = S_N(g)
	g = [g(i) for i in range(1, n+1)]
	H = SignedPermutations(n)
	return H(g)
	
def matrix_to_signed_permutation(n, M):
	sigma = [None for n in range(n)]
	for i in range(n):
		for j in range(n):
			if M[i][j] != 0:
				sigma[j] = (i+1)*M[i][j]
	return SignedPermutations(n)(sigma)

def signed_permutation_to_cycles(n, signed_permutation, signed=True):
	g = list(signed_permutation)
	if signed:
		S = list(range(-n, 0)) + list(range(1, n+1))
	else:
		S = list(range(1, n+1))
	S_N = SymmetricGroup(S)
	g = [g[i-1] for i in range(1,n+1)]
	if signed:
		g = list(reversed(list(map(lambda x: -1*x, g)))) + g
	return S_N(g)	
	
def signed_cycle_to_unsigned(n, disjoint_cycle_string): # Need to test this
	ans = []
	for a in disjoint_cycle_string.split(')('):
		ans.append((a.strip('(').strip(')')).split(','))
	solution = []
	return ans



## Models 

from enum import Enum, auto
from sage.all_cmdline import *
from cgt import conversions, helper_functions

class MODEL(Enum):
	TWO_REGION_INVERSIONS = auto()

def model(G, n, signed=True, model_type = MODEL.TWO_REGION_INVERSIONS):
	print("WARNING: only working for sage 'SymmetricGroup' G (signed or unsigned)")
	if model_type == MODEL.TWO_REGION_INVERSIONS:
		return __all_inversions_model(G, n, signed, num_regions = 2)
	
def __all_inversions_model(G, n, signed, num_regions):
	if num_regions == 2:
		if signed:
			return [G(f'({i},{-1*(i+1)})({-1*i},{i+1})') for i in range(1,n)] + [G(f'({n},-1)(-{n},1)')]
		else:
			return [G(f'({i},{i+1})') for i in range(1,n)] + [G(f'(1,{n})')]